#!/usr/bin/env python

# TODO
# resize when compose window open causes crash

import curses, textpad
import tpad
import select, sys, time, string, getopt, whrandom, re, os, operator
import cStringIO, linecache, signal
import gzip, cPickle
import termios, struct, fcntl

from types import *
from pygale import *

SEND_HEIGHT = 6
SCROLLBACK = 1000
CONNECTGALE = 1
VERSION_STRING = None	# must be overriden by vtfugu script
STATUS_TIMEOUT = 60 * 1000 	# in ms
DEBUG = 0
RECEIPT_TEXT = 'in.vtfugu'
USE_UNDERLINE = 1
PRIVMSG_BEEP = 1
SEND_IDLE = 1
CONFFILE = None	# default to ~/.gale/conf and vtfugustate.pck
VERBOSE_STATUS = 1

COLORS = {
	'red':	curses.COLOR_RED,
	'green':	curses.COLOR_GREEN,
	'yellow':	curses.COLOR_YELLOW,
	'blue':	curses.COLOR_BLUE,
	'magenta':	curses.COLOR_MAGENTA,
	'cyan':	curses.COLOR_CYAN,
	'white':	curses.COLOR_WHITE
}

def idletime(t):
	# t must be an integral number of seconds
	if t < 60:
		if t == 1:
			return '1s'
		else:
			return '%is' % t
	elif t < 3600:
		# up to an hour
		min = int(t/60.0)
		return '%im' % min
	elif t < 3600*24:
		# up to 24 hours
		return '%0.1fh' % (t/3600.0)
	else:
		days = int(t/(3600*24.0))
		return '%id' % days

def formatWithCharset(s):
	charset = pygale.gale_env.get('GALE_CHARSET',
		pygale.gale_env.get('CHARSET', 'ascii'))
	try:
		return s.encode(charset, 'replace')
	except LookupError:
		# quietly default to ascii if specified charset doesn't work
		return s.encode('ascii', 'replace')
	except UnicodeError:
		# default to ascii on unicode error (should not happen!)
		return s.encode('ascii', 'replace')

class PuffSender:
	def __init__(self, master, loc, body, callback):
		# callback is called with 1 on failure, 0 on success
		self.master = master
		list = string.split(loc)
		locs = filter(lambda x: x and x[0] != '/', list)
		locs = map(pygale.expand_aliases, locs)
		# filter out duplicates
		self.locs = []
		for loc in locs:
			if loc not in self.locs:
				self.locs.append(loc)

		# extract keywords
		self.keywords = filter(lambda x: x and x[0] == '/', list)
		self.keywords = map(lambda x: x[1:], self.keywords)
		self.body = body
		self.callback = callback

	def send(self):
		pygale.lookup_all_locations(self.locs, self.send2)
	
	def send2(self, loclist):
		send_locs = []
		bad_locs = []
		encr_recipients = []
		for (loc, recps) in loclist:
			if recps is None:
				bad_locs.append(loc)
			else:
				send_locs.append(loc)
				encr_recipients = encr_recipients + recps
		if bad_locs:
			s = 'Error: invalid location'
			if len(bad_locs) > 1:
				s = s + 's'
			s = s + ' ' + string.join(bad_locs)
			self.master.status.set(s)
			self.callback(1)
			return
		if '' in encr_recipients:
			# null key
			encr_recipients = []

		# Construct puff
		signer = pygale.gale_user()
		puff = pygale.Puff()
		gfrom = gale_env.get('GALE_NAME', gale_env.get('GALE_FROM',
			'vtFugu user'))
		puff.set_loc(string.join(send_locs, ' '))
		puff.set_text('message/sender', gfrom)
		puff.set_time('id/time', int(time.time()))
		puff.set_text('id/class', VERSION_STRING)
		puff.set_text('id/instance', pygale.getinstance())

		# Ask for a return receipt only if the message is encrypted and
		# our id is not in the list of encryption recipients
		private = encr_recipients and\
			not authcache.have_a_privkey(encr_recipients)
		if private:
			puff.set_text('question.receipt', pygale.gale_user())

		if self.keywords:
			for keyw in self.keywords:
				puff.set_text('message.keyword', keyw)
		puff.set_text('message/body', self.body)
		self.unsignedpuff = puff

		# Sign
		ret = puff.sign_message(signer)
		if ret is None:
			# unable to sign
			self.signed_puff = puff
		else:
			self.signed_puff = ret

		# Encrypt
		if encr_recipients:
			self.signed_puff.encrypt_message(encr_recipients,
				self.send3)
		else:
			self.send3(None)

	def send3(self, encrypted_puff):
		if encrypted_puff is None:
			# unable to encrypt
			encrypted_puff = self.signed_puff

		# echo to screen
		self.master.puffscreen.addpuff(self.unsignedpuff)

		self.master.conn.transmit_puff(encrypted_puff)
		self.callback(0)

# A status bar widget
class StatusBar:
	def __init__(self, win, master):
		self.create(win, master)

	def create(self, win, master):
		self.master = master
		self.statuswin = win
		(self.rows, self.cols) = self.statuswin.getmaxyx()
		self.statuswin.bkgd(curses.A_REVERSE)
		self.statuswin.erase()
		self.statuswin.keypad(1)		# enable arrow keys
		self.statuswin.noutrefresh()

	def clearTimeout(self):
		engine.engine.del_timeout(self.handle)

	def set(self, str):
		s = self.master.puffscreen.getPercentage()

		if len(str) < (self.cols - len(s)):
			str = str + ' ' * (self.cols - len(str) - 1)
		str = str[:self.cols-1-len(s)] + s
		if len(str) >= self.cols:
			str = str[:self.cols]
		if type(str) is UnicodeType:
			LOG('Converting unicode to ascii in status.set: %s' % `str`)
			str = formatWithCharset(str)
		if type(str) is not StringType:
			str = 'error: type %s' % type(str)
		self.statuswin.addstr(0, 0, str)
		self.statuswin.noutrefresh()
		self.handle = engine.engine.add_timeout(
			STATUS_TIMEOUT, self.erase)
		self.master.updateFocus()

	def erase(self):
		s = self.master.puffscreen.getPercentage()
		self.statuswin.erase()
		s = (' ' * (self.cols - 1 - len(s))) + s
		self.statuswin.addstr(0, 0, s)
		self.statuswin.noutrefresh()
		self.master.updateFocus()
	
	def moveTo(self, row):
		self.statuswin.mvwin(row, 0)
		self.statuswin.noutrefresh()

	def focus(self):
		self.statuswin.move(0, self.cols-1)
		self.statuswin.noutrefresh()

	def redraw(self):
		s = self.master.puffscreen.getPercentage()
		self.statuswin.addstr(0, self.cols-1-len(s), s)
		self.focus()

	def askchar(self, msg):
		self.statuswin.erase()
		self.statuswin.addstr(0, 0, msg[:self.cols-1])
		ch = self.statuswin.getch()
		self.statuswin.erase()
		self.redraw()
		return ch

	def askstring(self, msg, default=None):
		if default:
			msg = msg + '[%s] ' % default
		self.statuswin.erase()
		self.statuswin.addstr(0, 0, msg[:self.cols-1])
		curses.echo()
		try:
			s = self.statuswin.getstr()
		except KeyboardInterrupt:
			s = None
		curses.noecho()
		self.statuswin.erase()
		self.redraw()
		if s is None or s != '':
			return s
		else:
			return default

def LOG(msg):
	if DEBUG:
		f = open('log', 'a')
		f.write('%s: %s\n' %
			(time.asctime(time.localtime(time.time())), msg))
		f.close()

class Compose:
	def __init__(self, win, parent):
		self.win = win
		self.editor = tpad.Tpad(win)
		tpad.setErrorHandler(parent.status.set)
		self.win.erase()
		self.win.noutrefresh()

	def focus(self):
		self.editor.focus()
		self.win.noutrefresh()

	def addch(self, ch):
		ret = self.editor.safeprocess(ch)
		self.win.noutrefresh()
		return ret

	def get(self):
		return re.sub('\n', '\r\n', string.rstrip(self.editor.get())
			+ '\n')

class TextInput:
	def __init__(self, win, prompt):
		self.win = win
		(self.rows, self.cols) = self.win.getmaxyx()
		self.prompt = prompt
		(y, x) = win.getbegyx()
		subwin = win.subwin(self.rows, self.cols-len(self.prompt),
			y, x+len(self.prompt))
		self.editor = textpad.Textbox(subwin)
		self.editor.stripspaces = 1
		self._create_widgets()
	
	def _create_widgets(self):
		self.win.addstr(0, 0, self.prompt)
		self.win.noutrefresh()
	
	def focus(self):
		self.editor.win.move(0, len(self.editor.gather()))
		self.editor.win.noutrefresh()
	
	def addch(self, ch):
		try:
			ret = self.editor.do_command(ch)
		except curses.error:
			ret = 1
		self.editor.win.noutrefresh()
		return ret

	def get(self):
		return self.editor.gather()
	
	def set(self, msg):
		rows, cols = self.editor.win.getmaxyx()
		if len(msg) >= cols - 1:
			msg = msg[:cols-1]
		self.editor.win.erase()
		self.editor.win.addstr(0, 0, msg)
		self.editor.win.noutrefresh()

class Screen:
	def __init__(self, master, height, colors):
		self.master = master
		(self.rows, self.cols) = self.master.mainwin.getmaxyx()
		self.colors = colors

		self.height = height

		global SCROLLBACK
		SCROLLBACK = max(SCROLLBACK, self.height)
		self.pad = curses.newpad(SCROLLBACK+1, self.cols)
		self.pad.scrollok(1)

		for i in range(SCROLLBACK):
			self.pad.addstr(i, 0, '~')

		self.topRow = 0	# index of top row to be displayed on screen
						# in pad coordinates

		self.currline = 0
		self.pufflist = []
		self.puffLineStart = 0	# master offset; all puff line coords are
								# relative to this index (?)
		self.puffCursor = None
		self.MYrefresh()
	
	def getPercentage(self):
		if self.currline <= self.height:
			# it's all visible
			return ' --%'
		percentage = int(100.0 * self.topRow / 
                        (self.currline - self.height))
		return '%3i%%' % percentage
	
	def recreate(self, rows, cols, height):
		self.rows = rows
		self.cols = cols
		self.height = height
		global SCROLLBACK
		SCROLLBACK = max(SCROLLBACK, self.height)
		self.pad = curses.newpad(SCROLLBACK+1, self.cols)
		self.pad.scrollok(1)
		self.setPuffList(self.pufflist)
		self.MYrefresh()
	
	def setPuffList(self, list):
		self.pufflist = []
		self.pad.erase()
		self.puffCursor = None
		self.topRow = 0
		self.puffLineStart = 0
		self.currline = 0
		for p in list:
			self.addpuff(p)
		if self.pufflist:
			self.highlightPuff(on=0)
			self.puffCursor = self.pufflist[-1]
			self.highlightPuff(on=1)
		self.MYrefresh()

	def setHeight(self, height):
		# change the height of the puff screen window
		oldbottom = self.topRow + self.height - 1
		self.height = height
		self.topRow = oldbottom - self.height + 1
		self.MYrefresh()
	
	def MYrefresh(self):
		# check for bottom falling off the end
		bottom = self.topRow + self.height - 1
		if bottom >= SCROLLBACK:
			bottom = SCROLLBACK - 1
		if self.pufflist:
			# if you haven't filled up the pad yet
			lastPuffLine = self.pufflist[-1].startOffset +\
				self.pufflist[-1].numLines - 1
			if bottom >= lastPuffLine:
				bottom = lastPuffLine
		# make positive if it's negative
		self.topRow = bottom - self.height + 1
		self.topRow = max(0, self.topRow)

		self.pad.noutrefresh(self.topRow,0, 0,0,
			self.height-1,self.cols)
		if hasattr(self.master, 'status'):
			self.master.status.redraw()

	def repaint(self):
		# this doesn't currently work
		self.pad.erase()
		self.currline = 0
		self.topRow = 0
		self.puffLineStart = 0
		for p in self.pufflist:
			self.drawpuff(p)

	def isReceipt(self, puff):
		# Receipt (new-style only)
		receipt = puff.get_text_first('answer.receipt', None)
		return receipt is not None

	def addpuff(self, p, new=0):
		# Detect return receipts first
		if self.isReceipt(p):
			self.master.status.set(self.formatReceipt(p))
			return

		if not p.get_text_first('message/body'):
			# don't display puffs with no message body
			return

		# determine whether or not to scroll
		bottomvisible = self.topRow + self.height
		if self.topRow <= self.currline - 1 <= bottomvisible:
			scroll = 1
		else:
			scroll = 0

		numlines = self.formatPuff(p, new)
		if not self.pufflist:
			p.startOffset = 0
		else:
			last = self.pufflist[-1]
			p.startOffset = last.startOffset + last.numLines
		p.numLines = numlines
		self.pufflist.append(p)

		# update self.topRow if we should scroll
		if scroll:
			self.topRow = self.currline - self.height

		if self.puffCursor is None and new:
			# if no puff is selected and this is a new incoming puff,
			# select it
			self.puffCursor = p
			self.highlightPuff(on=1)
		self.MYrefresh()

	def formatPuff(self, puff, new=0):
		numlines = 0
		termwidth = self.cols
		#--------------------------------------------------
		# Catch all the message headers
		#--------------------------------------------------
		# Sender: real name of person sending puff
		sender = puff.get_text_first('message/sender', 'Unknown sender')
		# Location, eg "pub.comp.linux@ofb.net"
		locs = puff.get_loc()
		# Sign: Gale ID of sender, eg "tlau@ofb.net"
		signer = puff.get_signer('Unknown sig')
		if signer is None:
			signer = '*unsigned*'
		# Date message was sent, eg "1998-08-24 15:18:47"
		date = puff.get_time_first('id/time', None)
		if date is None:
			date = time.strftime('%H:%M:%S',
				time.localtime(time.time()))
		else:
			date = time.strftime('%H:%M:%S', time.localtime(date))
		# Encryption recipients, list of ids
		recipients = puff.get_recipients()
		if '' in recipients:
			# null key
			recipients = []
		# Status of recipient for a receipt, eg "in/present"
		status = puff.get_text_first('notice/presence', None)
		# Receipt (new-style only)
		receipt = puff.get_text_first('answer.receipt', None)
		# Client used to send puff
		idclass = puff.get_text_first('id/class', 'Unknown client')

		# Keywords
		keywords = puff.get_text('message.keyword')
		if keywords:
			keyw_text = map(lambda x: '/' + x, keywords)
			locs = locs + ' %s' % string.join(keyw_text, ' ')

		# presence
		presences = puff.get_text('notice/presence')
		if presences:
			locs = locs + ' %s' % string.join(presences, ' ')

		# Get the text of the message
		text = puff.get_text_first('message/body', '')
	
		# Beep on "private" puff
		# Private means it was encrypted but not to the signer
		if recipients and signer and signer not in recipients:
			if new and PRIVMSG_BEEP:
				curses.beep()

		# Hide puffs from certain authors
		hide_senders = pygale.gale_env.get('GALE_HIDE_SENDER', '')
		hide = 0
		if hide_senders:
			hide_senders = string.split(hide_senders)
			if puff.get_signer(None) in hide_senders:
				hide = 1

		hide_keywords = pygale.gale_env.get('GALE_HIDE_KEYWORDS', '')
		if hide_keywords:
			hide_keywords = string.split(hide_keywords)
			for word in hide_keywords:
				for kword in keywords:
					if kword.startswith(word):
						hide = 1
						break

		# Bold header
		s = ' %s %s' % (signer, date)
		if puff.get_text('question.receipt'):
			s = ' ?' + s
		if hide:
			s = ' X' + s
		s = formatWithCharset(s)
		loclen = termwidth - len(s)
		headerAttribs = curses.A_BOLD

		# colorize by signer
		if self.colors.has_key(signer):
			headerAttribs = headerAttribs | \
				curses.color_pair(self.colors[signer])

		locs = formatWithCharset(locs)
		if len(locs) > loclen:
			loclist = string.split(prettypuff.format(locs, loclen), '\n')
			if not loclist[-1]:
				# strip trailing newline
				loclist = loclist[:-1]
			s = loclist[0] + ' ' * (termwidth-len(s)-len(loclist[0])) + s
			self.addline(s, headerAttribs)
			numlines = numlines + 1
			for loc in loclist[1:-1]:
				self.addline(loc, headerAttribs)
				numlines = numlines + 1
			last_s = loclist[-1]
			# pad out the string with space so the underline goes
			# all the way to the right edge
			last_s = last_s + ' ' * (termwidth - len(last_s))
			if (not text or hide) and USE_UNDERLINE:
				headerAttribs = headerAttribs | curses.A_UNDERLINE
			self.addline(last_s, headerAttribs)
			LOG('adding space-padded line: %s' % `last_s`)
			numlines = numlines + 1
		else:
			# single-line header
			s = locs + ' ' * (termwidth-len(s)-len(locs)) + s
			if (not text or hide) and USE_UNDERLINE:
				# puff header is only content
				headerAttribs = headerAttribs | curses.A_UNDERLINE
			self.addline(s, headerAttribs)
			numlines = numlines + 1
		puff.header = s

		# Format message
		if text and not hide:
			text = formatWithCharset(text)
			text = string.join(map(self.esc, text), '')
			lines = string.split(
				prettypuff.format(text, termwidth), '\n')
			if not lines[-1]:
				# strip trailing blank
				lines = lines[:-1]
			for line in lines[:-1]:
				self.addline(line)
				numlines = numlines + 1
			lines[-1] = lines[-1] + ' ' * (termwidth-len(lines[-1]))
			if USE_UNDERLINE:
				attr = curses.A_UNDERLINE
			else:
				attr = 0
			self.addline(lines[-1], attr)
			numlines = numlines + 1
		return numlines

	def esc(self, x):
		if ord(x) < 0x20 and x not in ['\n', '\r', '\t']:
			return '^' + chr(ord(x) + 64)
		else:
			return x

	def formatReceipt(self, puff):
		#--------------------------------------------------
		# Catch all the message headers in the envvars
		#--------------------------------------------------
		# Sender: real name of person sending puff
		sender = puff.get_text_first('message/sender', 'Unknown sender')
		# Sign: Gale ID of sender, eg "tlau@ofb.net"
		signer = puff.get_signer('Unknown sig')
		if signer is None:
			signer = '*unsigned*'
		# Date message was sent, eg "1998-08-24 15:18:47"
		date = puff.get_time_first('id/time', None)
		if date is None:
			date = time.strftime('%m-%d %H:%M:%S',
				time.localtime(time.time()))
		else:
			date = time.strftime('%m-%d %H:%M:%S', time.localtime(date))
		# Status of recipient for a receipt, eg "in/present"
		status = puff.get_text_first('notice/presence', '???')

		# Format a return receipt
		s = 'Received: %s %s (%s) %s' % (
			status, signer, sender, date)

		s = formatWithCharset(s)

		return s

	def addFake(self):
		p = pygale.Puff()
		words = []
		for i in range(whrandom.randint(1, 5)):
			r = whrandom.randint(0, 45391)
			words.append(string.strip(
				linecache.getline('/usr/share/dict/words', r)))
		p.set_text('message/body', string.join(words, '\r\n'))
		p.set_time('id/time', int(time.time()))
		gfrom = gale_env.get('GALE_NAME', gale_env.get('GALE_FROM',
			'vtFugu user'))
		p.set_text('message/sender', gfrom)
		p.set_loc('fake@ofb.net')
		self.addpuff(p)

	def addline(self, line, attribs=0):
		if self.currline >= SCROLLBACK:
			# scroll everything up by one
			self.pad.scroll()
			self.currline = SCROLLBACK-1
			# update topRow to keep the same view on screen
			self.topRow = self.topRow - 1

			# maintain the master offset
			self.puffLineStart = self.puffLineStart + 1

			# remove puffs scrolled off top of screen
			while self.pufflist and\
				self.pufflist[0].startOffset < self.puffLineStart:
				if self.puffCursor == self.pufflist[0]:
					if len(self.pufflist) > 1:
						self.puffCursor = self.pufflist[1]
					else:
						self.puffCursor = None
				del self.pufflist[0]
#			self.highlightPuff(on=1)

		self.pad.addstr(self.currline, 0, ' ' * (self.cols - 1))
		self.pad.addstr(self.currline, 0, line, attribs)
		self.currline = self.currline + 1

	def page_up(self):
		self.topRow = self.topRow - self.height + 1
		self.MYrefresh()

	def page_down(self):
		self.topRow = self.topRow + self.height - 1
		self.MYrefresh()

	def show_top(self):
		# show the top of the puff?
		self.topRow = 0
		self.MYrefresh()

	def show_bottom(self):
		# what does this do?
		self.topRow = SCROLLBACK
		self.MYrefresh()

	def line_up(self):
		# not currently bound to any key
		self.topRow = self.topRow - 1
		self.MYrefresh()

	def line_down(self):
		# not currently bound to any key
		self.topRow = self.topRow + 1
		self.MYrefresh()

	def puff_up(self):
		if self.puffCursor is None:
			if self.pufflist:
				self.puffCursor = self.pufflist[0]
			return
		i = self.pufflist.index(self.puffCursor)
		if i == 0:
			self.show_top()
			return

		self.selectPuff(self.pufflist[i-1])

	def puff_down(self):
		if self.puffCursor is None:
			if self.pufflist:
				self.puffCursor = self.pufflist[-1]
			return
		i = self.pufflist.index(self.puffCursor)
		if i == len(self.pufflist) - 1:
			self.show_bottom()
			return

		self.selectPuff(self.pufflist[i+1])
	
	def selectPuff(self, puff):
		# return -1 on error (puff not in pufflist)
		# or 0 on success
		if puff not in self.pufflist:
			return -1
		self.highlightPuff(on=0)
		self.puffCursor = puff
		self.highlightPuff(on=1)

		# scroll to this puff
		puffoffset = self.puffCursor.startOffset - self.puffLineStart
		if puffoffset + self.puffCursor.numLines > \
			self.topRow + self.height:
			# it's off the bottom
			self.topRow = puffoffset + self.puffCursor.numLines -\
				self.height
		if puffoffset < self.topRow:
			# it's off the top
			self.topRow = puffoffset

		self.MYrefresh()
		return 0
	
	def show_details(self):
		if not self.puffCursor: return
		sender = self.puffCursor.get_text_first('message/sender', '???')
		client = self.puffCursor.get_text_first('id/class', '???')
		txt = '%s | %s' % (sender, client)
		txt = formatWithCharset(txt)
		self.master.status.set(txt)

	def select_first(self):
		if not self.pufflist:
			return
		self.selectPuff(self.pufflist[0])
		self.show_top()

	def select_last(self):
		if not self.pufflist:
			return
		self.selectPuff(self.pufflist[-1])
		self.show_bottom()
	
	def select_top(self):
		# select the topmost puff visible on the screen
		if not self.pufflist:
			return
		# find the first puff that starts after top
		for p in self.pufflist:
			offset = p.startOffset - self.puffLineStart
			if offset >= self.topRow:
				self.selectPuff(p)
				return

	def select_percentage(self, perc):
		# perc is a fraction down the screen; 0.0 means the first line
		# on screen, 1.0 means the last.
		# Select the last puff whose first line begins before the
		# designated line on screen.
		if not self.pufflist: return
		top = self.topRow
		bottom = self.topRow + self.height - 1
		target = top + (bottom - top) * perc
		for i in range(len(self.pufflist)):
			p = self.pufflist[i]
			pufftarget = p.startOffset - self.puffLineStart
			if pufftarget > target:
				if i == 0:
					self.selectPuff(self.pufflist[i])
				else:
					self.selectPuff(self.pufflist[i-1])
				return

		# if we get here, we didn't find anything
		# select the last puff in the buffer
		self.selectPuff(self.pufflist[-1])

	def _category_regexp(self, locstr):
		# create a regexp to recognize any location in locstr
		# this breaks on locations containing spaces
		# Split locstr into list of locations
		locs = string.split(locstr, None)
		locs = map(re.escape, locs)
		return re.compile(string.join(locs, '|'))

	def location_search(self, dir):
		if self.puffCursor is None: return
		orig_re = self._category_regexp(self.puffCursor.get_loc())
		func = lambda p, r=orig_re, s=self: \
			filter(None, map(r.match, p.get_loc().split(None)))
		self.master.lastSearch = func
		self.master.lastSearchDir = dir
		
		self.generic_search(dir, func)

	def generic_search(self, dir, func):
		# dir 0 means search forward
		# dir -1 means search backward
		if not self.puffCursor:
			puffindex = 0
		else:
			puffindex = self.pufflist.index(self.puffCursor)
		if dir == 0:
			list = range(puffindex+1, len(self.pufflist))
		elif dir == -1:
			list = range(puffindex-1, -1, -1)
		for i in list:
			if func(self.pufflist[i]):
				self.selectPuff(self.pufflist[i])
				return
		if dir == 0:
			self.master.status.set('No further matching puffs')
		elif dir == -1:
			self.master.status.set('No previous matching puffs')

	# Find the next puff by this author.  Do not wrap around.
	def author_search(self, dir):
		# dir 0 means search forward
		# dir -1 means search backward
		if self.puffCursor is None: return
		sender = self.puffCursor.get_text_first( 'message/sender', None)
		signer = self.puffCursor.get_signer(None)

		func = lambda p, sender=sender, signer=signer: (
			(sender and sender == p.get_text_first('message/sender', None))
			or
			(signer and signer == p.get_signer(None))
			)
		self.master.lastSearch = func
		self.master.lastSearchDir = dir

		self.generic_search(dir, func)

	def text_search(self, s, dir=0):
		# dir 0 means search forward
		# dir -1 means search backward
		if not self.pufflist: return

		func = lambda p, s=s: (
			(p.get_loc() and p.get_loc().find(s) != -1) or
			(p.get_signer(None) and p.get_signer().find(s) != -1) or
			(p.get_text_first('message/body') and
				p.get_text_first('message/body').find(s) != -1)
			)
		self.master.lastSearch = func
		self.master.lastSearchDir = dir

		self.generic_search(dir, func)

	def highlightPuff(self, on=0):
		# highlight currently selected puff
		if not self.puffCursor: return

		if on:
			# huyle complained about A_BOLD being hard to read in
			# reverse video
			# attr = curses.A_REVERSE | curses.A_BOLD
			attr = curses.A_REVERSE
		else:
			attr = curses.A_BOLD
		signer = self.puffCursor.get_signer()
		if self.colors.has_key(signer):
			attr = attr | curses.color_pair(self.colors[signer])

		if self.puffCursor.numLines == 1:
			attr = attr | curses.A_UNDERLINE
		self.pad.addstr(
			self.puffCursor.startOffset-self.puffLineStart,
			0,
			self.puffCursor.header, attr)

class UI:
	def __init__(self, root=None, subs=None, colors=None):
		if root is None:
			self.mainwin = curses.initscr()
		else:
			self.mainwin = root
		self.subs = subs
		self.colors = colors
		curses.cbreak()
		curses.noecho()
		curses.raw()						# enable raw mode
		curses.meta(1)						# enable 8-bit input
		(self.rows, self.cols) = self.mainwin.getmaxyx()

		self._create_widgets()
		self.focus = 0
		self.locbox = None
		self.compose = None
		self.marks = {}

		self.lastSearch = None			# function to repeat last search
		self.lastSearchDir = None		# last search direction
		self.lastSearchText = None		# text of last text search

		# load up stored puffs before connecting to galed
		self.loadPufflog()

		if CONNECTGALE: self.initGale()

		signal.signal(signal.SIGWINCH, self.resize)
		signal.signal(signal.SIGHUP, self.quit)
		signal.signal(signal.SIGTERM, self.quit)
		self.lastKeyPress = time.time()
		if VERBOSE_STATUS:
			self.status.set('Welcome to vtFugu')

	def updateFocus(self):
		if self.focus == 0:
			# focus in main window
			self.status.focus()
		elif self.focus == 1:
			# focus in locbox
			self.locbox.focus()
		elif self.focus == 2:
			self.compose.focus()
	
	def resize(self, signum, stackframe):
		self.rows, self.cols = struct.unpack(
			'4h', fcntl.ioctl(sys.stdout.fileno(),
			termios.TIOCGWINSZ,
			struct.pack('4h', 0, 0, 0, 0)))[:2]
		LOG('new height is %i' % self.rows)
		curses.endwin()
		self.mainwin = curses.initscr()
		curses.cbreak()
		curses.noecho()
		curses.raw()						# enable raw mode

		# recreate status bar
		self.status.create(curses.newwin(1, self.cols,
			self.rows - 1, 0), self)

		self.puffscreen.recreate(self.rows, self.cols, self.rows - 1)

		curses.doupdate()

	def initGale(self):
		if VERBOSE_STATUS:
			pygale.set_update_handler(self.status.set)
		pygale.set_error_handler(self.status.set)
		self.conn = pygale.GaleClient()
		self.conn.set_onconnect(self.onConnect)
		self.conn.connect(self.connected)

	def connected(self, host):
		if host is None:
			self.status.set('Unable to connect to a gale server')

	def onConnect(self, host):
		if VERBOSE_STATUS:
			self.status.set('Connected to gale server %s' % host)

		self.subs = map(pygale.expand_aliases, self.subs)
		if not self.subs:
			self.subs = [pygale.gale_user()]

		# listen for AKD requests
		akd_cat = '_gale.query.' + pygale.gale_user()
		if akd_cat not in self.subs:
			self.subs.insert(0, akd_cat)
		self.conn.sub_to(self.subs, self.subbed)

	def subbed(self, bad, good):
		if bad:
			for b in bad:
				self.status.set('Unable to find location %s' %
					formatWithCharset(b))
		if not good:
			self.status.set('No locations found')
		elif VERBOSE_STATUS:
			self.status.set('Subscribed to %s' %
				formatWithCharset(string.join(good, ' ')))
		self.conn.set_puff_callback(self.recvpuff)

	def _create_widgets(self):
		self.puffscreen = Screen(self, self.rows-1, self.colors)
		# Divider between puffscreen and composition widgets
		comp_row = self.rows - 1
		self.status = StatusBar(curses.newwin(1, self.cols,
			comp_row, 0), self)

	def setReadMark(self):
		self.marks[ord("'")] = self.puffscreen.puffCursor

	def processKey(self):
		self.lastKeyPress = time.time()
		ch = self.status.statuswin.getch()
		if self.focus == 0:
			# focus in main window
			if ch > 255:
				# special key
				key = curses.keyname(ch)
				if key == 'KEY_PPAGE':
					self.puffscreen.page_up()
					self.setReadMark()
				elif key == 'KEY_NPAGE':
					self.puffscreen.page_down()
					self.setReadMark()
				elif key == 'KEY_UP':
					self.puffscreen.puff_up()
					self.setReadMark()
				elif key == 'KEY_DOWN':
					self.puffscreen.puff_down()
					self.setReadMark()
				return

			key = curses.unctrl(ch)
			if key == 'q':
				# Quit
				self.quit()
			elif key == '^B' or key == '^H' or key == '^?' or key == 'b':
				self.puffscreen.page_up()
				self.setReadMark()
			elif key == '^F' or key == ' ':
				self.puffscreen.page_down()
				self.setReadMark()
			elif key == 'k':
				self.puffscreen.puff_up()
				self.setReadMark()
			elif key == 'j':
				self.puffscreen.puff_down()
				self.setReadMark()
			elif key == '^C':
				self.popdownCompose()
			elif key == '^I' or key == 'c':
				if self.locbox is not None:
					self.focus = 1
					self.locbox.focus()
				else:
					self.popupCompose()
			elif key == 'f' and DEBUG:
				self.puffscreen.addFake()
			elif key == '^R' or key == 'r':
				self.reply()
			elif key == '^E' or key == 'e':
				self.reply_to_sender()
			elif key == 'g':
				self.puffscreen.select_first()
				self.setReadMark()
			elif key == 'G':
				self.puffscreen.select_last()
				self.setReadMark()
			elif key == '^P':
				self.puffscreen.location_search(dir=-1)
				self.setReadMark()
			elif key == '^N':
				self.puffscreen.location_search(dir=0)
				self.setReadMark()
			elif key == 'a' or key == 'A':
					if key == 'a':
						dir = 0
					else:
						dir = -1
					self.puffscreen.author_search(dir=dir)
					self.setReadMark()
			elif key == 'H':
				self.puffscreen.select_top()
				self.setReadMark()
			elif key == 'M':
				self.puffscreen.select_percentage(0.5)
				self.setReadMark()
			elif key == 'L':
				self.puffscreen.select_percentage(1.0)
				self.setReadMark()
			elif key == '%':
				self.puffscreen.show_details()
			elif key == 'h':
				pass
				# reserved for help
			elif key == '/':
				# search forward
				s = self.status.askstring('Search forward: ',
					self.lastSearchText)
				if s:
					self.puffscreen.text_search(s, dir=0)
					self.lastSearchText = s
			elif key == '?':
				# search backward
				s = self.status.askstring('Search backward: ',
					self.lastSearchText)
				if s:
					self.puffscreen.text_search(s, dir=-1)
					self.lastSearchText = s
			elif key == 'n' or key == 'N':
				# search using last query
				if self.lastSearch:
					# forward or reverse
					if key == 'n':
						dir = self.lastSearchDir
					else:
						dir = -1-self.lastSearchDir
					self.puffscreen.generic_search(dir, self.lastSearch)
			elif key == 'm':
				if not self.puffscreen.puffCursor:
					self.status.set('No puff selected')
				else:
					mark = self.status.askchar('Set mark letter: ')
					LOG('mark: %s' % mark)
					self.marks[mark] = self.puffscreen.puffCursor
			elif key == "'":
				mark = self.status.askchar('Go to mark: ')
				if self.marks.has_key(mark):
					ret = self.puffscreen.selectPuff(self.marks[mark])
					if ret == -1:
						self.status.set('Mark no longer in scrollback')
				else:
					self.status.set('Unknown mark')
			else:
				pass
		elif self.focus == 1:
			# focus is in location box
			# process character
			key = curses.keyname(ch)
			if key == '^C':
				# cancel send
				self.popdownCompose()
				return
			elif key == '^X':
				# Pine binding courtesy of Huy Le
				self.send()
				return
			elif key == '^I':
				# focus on textbox
				self.compose.focus()
				self.focus = 2
				return

			# otherwise, pass it through to the return window
			ret = self.locbox.addch(ch)
			if not ret:
				# terminated
				# change focus
				self.compose.focus()
				self.focus = 2
		elif self.focus == 2:
			# process character
			key = curses.keyname(ch)
			if key == '^C':
				# cancel send
				self.popdownCompose()
			elif key == '^I':
				self.status.focus()
				self.focus = 0
			elif key == '^X':
				self.send()
			else:
				# add it to the compose window
        			ret = self.compose.addch(ch)
				if ret:
					self.send()

	def send(self):
		# user hit a key that initiates sending the puff
		self.status.focus()
		self.focus = 0
		self.sendpuff(self.locbox.get(),
			self.compose.get())

	def reply(self):
		if self.puffscreen.puffCursor is None:
			return
		p = self.puffscreen.puffCursor

		# Figure out some nice defaults
		# Check for login/logout notices
		notice_re = re.compile('_gale\.notice\.(.*)')
		notice_groups = notice_re.match(p.get_loc())

		dests = []
		recipients = p.get_recipients()
		signer = p.get_signer(None)

		if notice_groups:
			# Handle notice messages separately
			dests = [notice_groups.group(1)]
		elif not recipients or signer in recipients:
			# Public unencrypted puff or group send
			dests = p.get_loc_list()
		else:
			dests = p.get_loc_list()
			if signer is not None and signer not in ['*verifying*',
				'*unverified*', '*unsigned*']:
				dests.append(signer)
		dests = filter(lambda x: x != pygale.gale_user(), dests)
		if not dests:
			# No one to send to!  Send to myself
			dests = [pygale.gale_user()]

		# Add in the keywords
		keywords = map(lambda x: '/' + x,
			p.get_text('message.keyword'))
		dests = dests + keywords

		self.popupCompose()
		self.locbox.set(formatWithCharset(string.join(dests, ' ')))
		self.focus = 2
		self.compose.focus()

	def reply_to_sender(self):
		p = self.puffscreen.puffCursor
		if p is None: return
		if p.get_signer(None) is None: return

		dests = [p.get_signer()]
		# Add in the keywords
		keywords = map(lambda x: '/' + x,
			p.get_text('message.keyword'))
		dests = dests + keywords

		self.popupCompose()
		self.locbox.set(formatWithCharset(string.join(dests, ' ')))
		self.focus = 2
		self.compose.focus()

	def popupCompose(self):
		self.puffscreen.setHeight(self.rows - SEND_HEIGHT)
		self.status.moveTo(self.rows - SEND_HEIGHT)
		self.locbox = TextInput(curses.newwin(1, self.cols,
			self.rows - SEND_HEIGHT + 1, 0), 'To: ')
		self.focus = 1
		self.compose = Compose(curses.newwin(SEND_HEIGHT - 2,
			self.cols, self.rows - SEND_HEIGHT + 2, 0), self)
		self.locbox.focus()

	def popdownCompose(self):
		self.status.moveTo(self.rows - 1)
		self.puffscreen.setHeight(self.rows - 1)
		self.mainwin.touchwin()
		self.locbox = None
		self.compose = None
		self.focus = 0

	def sendpuff(self, loc, body):
		p = PuffSender(self, loc, body, self.puffsent)
		p.send()

	def puffsent(self, ret):
		if ret == 0:
			# success
			self.popdownCompose()
	
	def loop(self):
		engine.engine.add_callback(sys.stdin, self.processKey)
		while 1:
			curses.doupdate()
			engine.engine.process()

	def quit(self, *args):
		LOG('quitting program')
		self.persistPufflog()
		curses.nocbreak()
		self.mainwin.keypad(0)
		curses.echo()
		curses.endwin()
		sys.exit(0)

	def loadPufflog(self):
		# Fugu directory
		fugudir = pygale.gale_env.get('FUGU_DIR',
			os.path.join(userinfo.home_dir, '.gale'))

		# Load puff history
		if CONFFILE:
			fname = CONFFILE
		else:
			fname = 'vtfugustate'
		gzfname = os.path.join(fugudir, fname + '.pck') + '.gz'
		if not os.path.exists(gzfname):
			# No file found
			LOG('file not found: %s' % gzfname)
			return

		try:
			f = gzip.GzipFile(gzfname, 'rb')
			pufflog = cPickle.load(f)
			f.close()
		except KeyboardInterrupt:
			print 'Keyboard interrupt caught; exiting...'
			self.quit()
		except Exception, e:
			# file exists, but can't be loaded
			LOG('Error loading history: %s' % `e`)
			return
		self.puffscreen.setPuffList(pufflog)
	
	def persistPufflog(self):
		# Fugu directory
		fugudir = pygale.gale_env.get('FUGU_DIR',
			os.path.join(userinfo.home_dir, '.gale'))

		# Save puff history to disk
		pufflog = self.puffscreen.pufflist
		if not pufflog:
			return
		if CONFFILE:
			fname = CONFFILE
		else:
			fname = 'vtfugustate'
		gzfname = os.path.join(fugudir, fname + '.pck.gz')
		try:
			f = gzip.GzipFile(gzfname, 'wb')
			os.chmod(gzfname, 0600)
			cPickle.dump(pufflog, f, 1)
			f.close()
		except:
			# can't save; ignore error for now
			pass

	# Send our key if it was requested
	def send_akd_response(self, reqpuff):
		reqkey = reqpuff.get_text_first('question.key')
		LOG('Received AKD request for %s' % reqkey)
		pygale.export_pubkey(reqkey,
			lambda e, s=self, r=reqkey: s.akd_request2(r, e))

	def akd_request2(self, keyname, export_key):
		if not export_key:
			return
		LOG('Sending AKD response for key %s' % keyname)
		p = pygale.Puff()
		p.set_binary('answer/key', export_key)
		p.set_text('id/instance', pygale.getinstance())
		p.set_text('id/class', VERSION_STRING)
		p.set_loc('_gale.key.' + keyname)
		try:
			self.conn.transmit_puff(p)
		except pygale.PyGaleErr, e:
			# server connection died
			# Assume retry is already in progress
			self.status.set(e)

	# What to do when we receive a puff
	def recvpuff(self, p):
		# First check for AKD requests
		if p.get_text_first('question.key'):
			self.send_akd_response(p)
			return

		# Filter by sender
		# Do this before sending a receipt, below
		kill_senders = pygale.gale_env.get('GALE_KILL_SENDER', '')
		if kill_senders:
			kill_senders = string.split(kill_senders)
			if p.get_signer(None) in kill_senders:
				return

		# Filter by keyword
		kill_keywords = pygale.gale_env.get('GALE_KILL_KEYWORDS', '')
		if kill_keywords:
			kill_keywords = string.split(kill_keywords)
			keywords = p.get_text('message.keyword')
			for word in kill_keywords:
				for kword in keywords:
					if kword.startswith(word):
						return

		# Send return receipts
		if p.get_text('question.receipt'):
			self.send_recpt(p)
	
		# if we get here, the puff should be displayed
		self.puffscreen.addpuff(p, new=1)

	# Send a return receipt if we're supposed to
	# This must only be called if the puff contains a "question.receipt"
	# fragment
	def send_recpt(self, puff):
		# the user doesn't want us to send receipts
		if RECEIPT_TEXT is None: return
		p = pygale.Puff()
		p.set_loc(puff.get_text_first('question.receipt'))
		p.set_text('answer.receipt', pygale.gale_user())

		# calculate idle time
		if SEND_IDLE:
			idle = int(time.time() - self.lastKeyPress)
			presence = '%s (idle %s)' % (RECEIPT_TEXT, idletime(idle))
			p.set_time('status.idle', int(self.lastKeyPress))
			p.set_text('status.presence', RECEIPT_TEXT)
		else:
			presence = RECEIPT_TEXT

		p.set_text('notice/presence', presence)
		gfrom = gale_env.get('GALE_NAME', gale_env.get('GALE_FROM',
			'vtFugu user'))
		p.set_text('message/sender', gfrom)
		p.set_text('id/class', VERSION_STRING)
		p.set_time('id/time', int(time.time()))
		p.set_text('id/instance', pygale.getinstance())
		# Sign return receipt
		p = p.sign_message(pygale.gale_user())
		# Encrypt if original message was signed
		old_signer = puff.get_signer(None)
		if old_signer is not None:
			p.encrypt_message(old_signer,
				lambda p, s=self, orig=p: s.send_recpt2(p, orig))
		else:
			self.send_recpt2(p, p)
	
	def send_recpt2(self, puff, origpuff):
		if puff is None:
			self.status.set('Unable to encrypt return receipt')
			puff = origpuff
		try:
			self.conn.transmit_puff(puff)
		except pygale.PyGaleErr, e:
			# server connection died
			# Assume retry is already in progress
			self.status.set(e)

def parseColors(root, colors):
	if colors is None or not curses.has_colors():
		return {}
	
	# initialize colors; assume that initscr()/start_color() have already
	# been called
	for colorname, number in COLORS.items():
		# TODO: curses assumes the background is black... ugh
		LOG('initting color pair %i with %i/black' % (number, number))
		curses.init_pair(number, number, curses.COLOR_BLACK)

	colors = re.sub('\s+', ' ', colors)
	colors = string.split(colors)
	if len(colors) % 2 != 0:
		root.addstr('Invalid GALE_COLOR_SENDERS specification\n')
		root.addstr('Expected an even number of arguments\n')
		root.refresh()
		time.sleep(5)
		return {}

	ret = {}
	sleep = 0
	for i in range(len(colors)/2):
		sender = colors[2*i]
		color = colors[2*i+1]
		if not COLORS.has_key(color):
			root.addstr('Invalid color %s in GALE_COLOR_SENDERS\n' %
				color)
			root.refresh()
			sleep = 1
			continue
		ret[sender] = COLORS[color]
	if sleep:
		time.sleep(5)
	return ret

def usage():
	print 'vtfugu [options] loc1 loc2 ... locN'
	print 'Options:'
	print '   -h          This help message'
	print '   -d          Enable debugging (to file "log")'
	print '   -s #        Set the size of the scrollback buffer (in lines)'
	print '   -p presence Set your presence string for return receipts'
	print '               use "None" to disable return receipt sending'
	print '   -u          Disable underline as puff separator'
	print '   -b          Disable beeping on receipt of a private message'
	print '   -i          Disable sending idle time on return receipts'
	print '   -e          Disable default subscriptions (GALE_SUBSCRIBE)'
	print '   -f fn.conf  Use configuration file fn.conf (see README)'
	print '   -q          Suppress Gale warnings and notices'
	print '"loc" is a Gale location, e.g. "pub@ofb.net"'
	print 'Your userid is always added to the locations specified on the'
	print 'command line.'

def main(root):
	LOG('=' * 30)
	root.move(0, 0)
	whrandom.seed(int(time.time()) % 256)
	opts, args = getopt.getopt(sys.argv[1:], 'qds:hp:ubief:')
	global CONNECTGALE, DEBUG, SCROLLBACK, RECEIPT_TEXT, USE_UNDERLINE
	global SEND_IDLE, CONFFILE, PRIVMSG_BEEP, VERBOSE_STATUS
	ignore_subs = 0
	conffile = None
	for opt, val in opts:
		if opt == '-q':
			VERBOSE_STATUS = 0
		elif opt == '-d':
			DEBUG = 1
			tpad.DEBUG = 1
		elif opt == '-s':
			SCROLLBACK = int(val)
		elif opt == '-h':
			curses.endwin()
			usage()
			sys.exit()
		elif opt == '-u':
			USE_UNDERLINE = 0
		elif opt == '-b':
			PRIVMSG_BEEP = 0
		elif opt == '-p':
			if string.lower(val) == 'none':
				RECEIPT_TEXT = None
			else:
				RECEIPT_TEXT = val
		elif opt == '-i':
			SEND_IDLE = 0
		elif opt == '-e':
			ignore_subs = 1
		elif opt == '-f':
			fugudir = pygale.gale_env.get('FUGU_DIR',
				os.path.join(userinfo.home_dir, '.gale'))
			CONFFILE = os.path.join(fugudir, val)
			if not os.path.exists(CONFFILE):
				root.addstr('Configuration file %s does not exist!\n' %
					CONFFILE)
				root.refresh()
				time.sleep(5)

	# initialize pygale before fetching subs
	if CONFFILE:
		pygale.init(os.path.join(os.environ['HOME'], '.gale',
			CONFFILE))
	else:
		pygale.init()
	sys.exitfunc = pygale.shutdown
 
	# fetch default subs from ~/.gale/conf
	if not ignore_subs:
		s = pygale.gale_env.get('GALE_SUBSCRIBE', pygale.gale_user())
		s = re.sub('\s+', ' ', s)
		s = s.replace(' - ', ' -')
		default_subs = string.split(s)
	else:
		default_subs = []

	# get sender colorization
	colors = parseColors(root,
		pygale.gale_env.get('GALE_COLOR_SENDERS', None))

	subs = default_subs + args
	root.clear()
	ui = UI(root=root, subs=subs, colors=colors)
	ui.loop()

if __name__ == '__main__':
	curses.wrapper(main)

# vim:noet:ts=4:sw=4
